import csv
import os
import re
from typing import Optional

import pandas as pd


class DataDescription:
    def __init__(self, name: str, description: str):
        self.name = name
        self.description = description


def check_if_can_create_file(output_file) -> Optional[str]:
    if not output_file.endswith((".csv", ".xlsx", ".json")):
        return "Only \".csv\", \".xlsx\", or \".json\" file types are supported."

    try:
        with open(output_file, "w"):
            pass
    except PermissionError:
        return f"Permission denied for creating output file: \"{output_file}\"."
    except Exception as e:
        return f"An error occurred: {e}"

    return None


def save_result(output_file: str, query_param_names: list[str], result_rows: [dict[str, str]]) -> Optional[str]:
    output_rows: [dict[str, Optional[str]]] = []

    for result_row in result_rows:
        output_row: dict[str, Optional[str]] = {}
        for name in query_param_names:
            value = result_row.get(name)

            if value:
                output_row[name] = value
            else:
                output_row[name] = None
        output_rows.append(output_row)

    try:
        with open(output_file, "w", newline=""):
            df = pd.DataFrame(output_rows)

            if output_file.endswith(".csv"):
                df.to_csv(output_file, index=False, columns=query_param_names)
            if output_file.endswith(".xlsx"):
                df.to_excel(output_file, index=False, columns=query_param_names)
            if output_file.endswith(".json"):
                df.to_json(output_file, orient='records', lines=False, indent=4)
        return None
    except Exception as e:
        return f"Failed to write result to output file {e}"


def setup_output_file(output_file: str) -> Optional[str]:
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    with open(output_file, "w", newline="") as file:
        csv.writer(file)

    return None


def find_matched_files(input_string: str) -> [Optional[list[str]], Optional[str]]:
    parts = re.split(r"[,\n]+", input_string)
    cleaned_parts = [part.strip() for part in parts if part.strip()]

    files = []

    for part in cleaned_parts:
        files += find_matched_files_for_pattern(part)

    if not files:
        return None, f"No file found given locations: {input_string}"

    return files, None


def find_matched_files_for_pattern(pattern) -> [str]:
    if os.path.isdir(pattern):
        return [os.path.join(pattern, f) for f in os.listdir(pattern) if os.path.isfile(os.path.join(pattern, f))]

    if os.path.isfile(pattern):
        return [pattern]

    directory, file_pattern = os.path.split(pattern)
    if not directory:
        directory = "."

    matched_files = []
    for filename in os.listdir(directory):
        if re.match(file_pattern, filename):
            matched_files.append(os.path.join(directory, filename))

    return matched_files



import subprocess

from lumei.query_param import CommandQueryParam


def process_commands(
        input_file_path: str,
        commands: list[CommandQueryParam],
) -> dict[str, any]:
    results: dict[str, any] = {}

    for command in commands:
        prepared_command = command.command.replace("%input_file_path%", input_file_path)
        result = subprocess.run(prepared_command, shell=True, capture_output=True, text=True)
        results[command.name] = result.stdout.removesuffix("\n")

    return results



from lumei.openai.agent import (
    Agent,
    create_agent,
)
from lumei.openai.file_search import (
    FileSearchQueryParam,
    file_search,
)
from lumei.openai.standalone_file_search import (
    FileSearchException,
    openai_file_search,
)

__all__ = [
    "Agent",
    "create_agent",
    "FileSearchQueryParam",
    "file_search",
    "FileSearchException",
    "openai_file_search",
]



from typing import Optional

from lumei import create_agent, FileSearchQueryParam, file_search
from lumei.openai.agent import default_model


class FileSearchException(Exception):
    pass


def openai_file_search(
        openai_api_key: str,
        input_files: list[str],
        file_search_query: dict[str, str],
        model: Optional[str] = default_model,
        assistant_id: Optional[str] = None,
) -> list[dict[str, str] | FileSearchException]:
    agent, error = create_agent(
        openai_api_key=openai_api_key,
        model=model,
        assistant_id=assistant_id,
    )

    if error:
        raise FileSearchException(f"File search query failed with error: {error}")

    params = []

    for key, value in file_search_query.items():
        params.append(
            FileSearchQueryParam(
                name=key,
                description=value,
            )
        )

    results: list[dict[str, str] | FileSearchException] = []

    for input_file in input_files:
        result, error = file_search(
            agent=agent,
            input_file_path=input_file,
            file_search_query_params=params
        )

        if error:
            results.append(FileSearchException(f"File search query failed with error: {error}"))
        else:
            results.append(result)

    return results



from datetime import datetime

from lumei.attribute_processor import get_attribute_results
from lumei.command_processor import process_commands
from lumei.file_manager import save_result, find_matched_files, setup_output_file, \
    check_if_can_create_file
from lumei.openai.agent import create_agent
from lumei.openai.file_search import file_search, FileSearchQueryParam
from lumei.query_param import parse_query_string, QueryParamAttribute, CommandQueryParam


def execute_query_and_store_results(
        openai_api_key: str,
        input_files: str,
        output_file: str,
        file_search_query_string: str
) -> int:
    error = check_if_can_create_file(output_file)

    if error:
        print(error)
        return 1

    print("Verified permission to create file.")

    query, error = parse_query_string(file_search_query_string)

    if error:
        print(error)
        return 1

    file_search_query = []
    attribute_query: dict[str, QueryParamAttribute] = {}
    command_query: list[CommandQueryParam] = []
    query_param_names: list[str] = []
    for query_param in query:
        query_param_names.append(query_param.name)
        if query_param.name and query_param.search:
            file_search_query.append(
                FileSearchQueryParam(
                    name=query_param.name,
                    description=query_param.search,
                )
            )
        elif query_param.name and query_param.attribute:
            attribute_query[query_param.name] = query_param.attribute
        elif query_param.name and query_param.command:
            command_query.append(CommandQueryParam(
                name=query_param.name,
                command=query_param.command,
            ))

    print("Parsed list of names and data descriptions for file search.")

    files, error = find_matched_files(input_files)

    if error:
        print(error)
        return 1

    print(f"Created list of files to process. Found {len(files)} files.")

    agent, error = create_agent(openai_api_key)

    if error:
        print(error)
        return 1

    print(f"Created agent with assistant id: {agent.assistant_id}.")

    print(f"Begun processing files.")

    results: [dict[str, str]] = []

    for file in files:
        start_time = datetime.now()
        result: dict[str, any] = {}
        file_search_results, error = file_search(agent, file, file_search_query)
        end_time = datetime.now()

        if file_search_results:
            result.update(file_search_results)

            command_results = process_commands(
                input_file_path=file,
                commands=command_query
            )

            result.update(command_results)

            attribute_results = get_attribute_results(
                attribute_query=attribute_query,
                input_file_path=file,
                start_time=start_time,
                end_time=end_time,
            )

            result.update(attribute_results)

            results.append(result)
            print(f"    {file} SUCCESS")
        elif error:
            print(f"    {file} FAILED: {error}")
        else:
            print(f"    {file} FAILED")

    print(f"Finished processing files.")

    error = setup_output_file(output_file)

    if error:
        print(error)
        return 1

    print(f"Created output file: {output_file}.")

    error = save_result(output_file, query_param_names, results)

    if error:
        print(error)
        return 1

    print(f"Wrote result to file: {output_file}.")

    return 0



from typing import Optional

from openai import OpenAI


class Agent:
    def __init__(
            self, assistant_id: str,
            openai_client: OpenAI
    ):
        self.assistant_id = assistant_id
        self.client = openai_client


default_instructions = """"
    You are a file search assistant.
    You are given a file and a list of data to extract from the data. For each data, the user will specify a name for the data and a description of what the data is.
    The list of data to extract is in the form of a JSON object. The key is the name for the data and value is the description for the data.
    All the data that you should extract is in the file. If the data is not in the file, then do not return any information for that data. Do not hallucinate.
    All responses should be returned in the format of a proper JSON object. The keys for the JSON are the name for the data and the value is the extracted data. There should be no key in the response that is not found in the JSON object that holds the list of data to extract.
    Please and thank you.
"""

default_model = "gpt-4o"


def create_agent(
        openai_api_key: str,
        instructions: Optional[str] = default_instructions,
        model: Optional[str] = default_model,
        assistant_id: Optional[str] = None,
) -> [Optional[Agent], Optional[str]]:
    try:
        openai_client = OpenAI(
            api_key=openai_api_key
        )
    except Exception as e:
        return None, f"Failed to create OpenAI client: {e}"

    if not assistant_id:
        try:
            assistant = openai_client.beta.assistants.create(
                name="File Search",
                instructions=instructions,
                model=model,
                tools=[
                    {
                        "type": "file_search"
                    }
                ],
            )
            new_assistant_id = assistant.id
        except Exception as e:
            return None, f"Failed to create agent: {e}"
    else:
        new_assistant_id = assistant_id

    return Agent(
        assistant_id=new_assistant_id,
        openai_client=openai_client,
    ), None



import argparse
import os

from lumei.query_executor import execute_query_and_store_results

parser = argparse.ArgumentParser(description='Lumei file processor')

parser.add_argument(
    "--input-files",
    default=None,
    required=True,
    help="Source files to process on. "
         "Multiple files can be provided and they are seperated by a comma `,` character. "
         "File inputs can be expressed as a path to a single file or a regex."
)
parser.add_argument(
    "--output-file",
    default=None,
    required=True,
    help="Path of the file that the results will be written to. "
         "Input must be a file path to a single file. "
         "Supported file formate are `.csv`, `.xlsx`, and `.json`."
)
parser.add_argument(
    "--openai-api-key",
    default=None,
    required=True,
    help="API key for OpenAI, necessary for file search functionalities. "
         "Key can be obtained from here https://platform.openai.com/account/api-keys.\n\n"
         "Alternative way to provide the API key is to set it as the `OPENAI_API_KEY` environment variable."
)
parser.add_argument(
    "--query",
    default=None,
    required=True,
    help="Name and description of data to search for. "
         "Input should be an array of JSON objects. "
         "`name` is the name of the data to search for. "
         "Name of the data will be the column name for the result dataset. "
         "`search` is the description of the data to search for."
)
args = parser.parse_args()


def main():
    if args.openai_api_key:
        openai_api_key = args.openai_api_key
    elif os.getenv("OPENAI_API_KEY"):
        openai_api_key = os.getenv("OPENAI_API_KEY")
    else:
        openai_api_key = None

    if not openai_api_key:
        print("OpenAI API key not provided.")
        exit(1)

    print("Starting file search process.")

    exit_code = execute_query_and_store_results(
        input_files=args.input_files,
        output_file=args.output_file,
        openai_api_key=openai_api_key,
        file_search_query_string=args.query,
    )

    exit(exit_code)



import ast
from enum import Enum
from typing import Optional


class QueryParamAttribute(str, Enum):
    FILE_PATH = "FILE_PATH"
    START_TIMESTAMP = "START_TIMESTAMP"
    END_TIMESTAMP = "END_TIMESTAMP"
    START_DATETIME = "START_DATETIME"
    END_DATETIME = "END_DATETIME"


class FileSearchQueryParam:
    def __init__(self, name: str, description: str):
        self.name = name
        self.description = description


class CommandQueryParam:
    def __init__(
            self,
            name: str,
            command: str
    ):
        self.name = name
        self.command = command


class QueryParam:
    def __init__(
            self, name: str,
            search: Optional[str],
            attribute: Optional[QueryParamAttribute],
            command: Optional[str],
    ):
        self.name = name
        self.search = search
        self.attribute = attribute
        self.command = command


def parse_query_string(query_string: str) -> [Optional[list[QueryParam]], Optional[str]]:
    try:
        query_objects = ast.literal_eval(query_string)
    except Exception as e:
        return None, f"Failed to parse query with string: {e}"

    if query_objects is list:
        return None, f"Query must be an array of objects, got string: {query_string}"

    seen_param_names = []

    query = []

    for query_object in query_objects:
        name = query_object.get("name")
        search = query_object.get("search")
        attribute = query_object.get("attribute")
        command = query_object.get("command")

        if not name:
            return None, f"Query parameter name is required: {query_string}"

        if name in seen_param_names:
            return None, f"Found duplicate query parameter name: {name}"

        seen_param_names.append(name)

        if search and not attribute and not command:
            query.append(
                QueryParam(
                    name=name,
                    search=search,
                    attribute=None,
                    command=None,
                )
            )
        elif not search and attribute and not command:
            query.append(
                QueryParam(
                    name=name,
                    search=None,
                    attribute=QueryParamAttribute(attribute),
                    command=None,
                )
            )
        elif not search and not attribute and command:
            query.append(
                QueryParam(
                    name=name,
                    search=None,
                    attribute=None,
                    command=command,
                )
            )
        else:
            return None, (f"Error, query parameter should only have one of `search`, `attribute`, or `command`: "
                          f"{query_object}")

    return query, None



import json
import re
from typing import Optional

from openai.types import FileObject
from openai.types.beta import Thread
from openai.types.beta.threads import Run

from lumei.openai.agent import Agent
from lumei.query_param import FileSearchQueryParam


def file_search(
        agent: Agent,
        input_file_path: str,
        file_search_query_params: list[FileSearchQueryParam]
) -> [Optional[dict[str, any]], Optional[str]]:
    vector_store_id: Optional[str] = None
    thread: Optional[Thread] = None
    run: Optional[Run] = None
    query = {}

    for param in file_search_query_params:
        query[param.name] = param.description

    try:
        file = agent.client.files.create(
            file=open(input_file_path, "rb"),
            purpose="assistants",
        )
    except Exception as e:
        return None, f"Failed to upload file: {input_file_path}. {e}"

    try:
        thread = agent.client.beta.threads.create(
            messages=[
                {
                    "role": "user",
                    "content": json.dumps(query),
                    "attachments": [
                        {
                            "file_id": file.id,
                            "tools": [
                                {
                                    "type": "file_search"
                                }
                            ]
                        }
                    ],
                }
            ]
        )

        vector_store_id = thread.tool_resources.file_search.vector_store_ids[0]
    except Exception as e:
        clean_up_request(agent, file, vector_store_id)
        return None, f"Failed to create thread or vector store: {thread}. {e}"

    try:
        run = agent.client.beta.threads.runs.create_and_poll(
            thread_id=thread.id,
            assistant_id=agent.assistant_id,
            temperature=0,
        )
    except Exception as e:
        clean_up_request(agent, file, vector_store_id)
        return None, f"Failed to process request with thread: {run}. {e}"

    try:
        messages = list(agent.client.beta.threads.messages.list(thread_id=thread.id, run_id=run.id))
        message_content = messages[0].content[0].text.value
    except Exception as e:
        clean_up_request(agent, file, vector_store_id)
        return None, f"Failed to extract result text from thread. {e}"

    clean_up_request(agent, file, vector_store_id)

    formatted_result = format_result(message_content)

    if formatted_result:
        return formatted_result, None
    else:
        return None, f"Failed to extract data from response: {formatted_result}"


def format_result(result_text: str) -> Optional[dict[str, any]]:
    match = re.search(r'\{.*}', result_text, re.DOTALL)
    if match:
        result_json = match.group()
        try:
            return json.loads(result_json)
        except json.JSONDecodeError:
            return None
    return None


def clean_up_request(agent: Agent, file: Optional[FileObject], vector_store_id: Optional[str]):
    if file:
        try:
            agent.client.files.delete(
                file_id=file.id
            )
        except Exception as e:
            print(f"Failed to delete file: {file.id}. Error: {e}")

    if vector_store_id:
        try:
            agent.client.beta.vector_stores.delete(
                vector_store_id=vector_store_id
            )
        except Exception as e:
            print(f"Failed to delete vector store: {vector_store_id}. Error: {e}")



from datetime import datetime

from lumei.query_param import QueryParamAttribute


def get_attribute_results(
        attribute_query: dict[str, QueryParamAttribute],
        input_file_path: str,
        start_time: datetime,
        end_time: datetime,
) -> dict[str, any]:
    result: dict[str, any] = {}

    for key, attribute in attribute_query.items():
        if attribute is QueryParamAttribute.FILE_PATH:
            result[key] = input_file_path
        elif attribute is QueryParamAttribute.START_TIMESTAMP:
            result[key] = start_time.timestamp()
        elif attribute is QueryParamAttribute.START_DATETIME:
            result[key] = start_time.strftime('%Y-%m-%d %H:%M:%S')
        elif attribute is QueryParamAttribute.END_TIMESTAMP:
            result[key] = end_time.timestamp()
        elif attribute is QueryParamAttribute.END_DATETIME:
            result[key] = end_time.strftime('%Y-%m-%d %H:%M:%S')

    return result



# from lumei.openai.agent import (
#     Agent,
#     create_agent,
# )
# from lumei.openai.file_search import (
#     FileSearchQueryParam,
#     file_search,
# )
#
# __all__ = [
#     "Agent",
#     "create_agent",
#     "FileSearchQueryParam",
#     "file_search",
# ]
